@@ -3,19 +3,21 @@ Tools for reading a nexus file
 """
 import re
 import os
+import gzip
 import warnings
 import pickle
 from hashlib import md5
 
-try:
+try:  # pragma: no cover
     from StringIO import StringIO
 except ImportError:
     from io import StringIO
 
 DEBUG = False
 
-BEGIN_PATTERN = re.compile(r"""begin (\w+);""", re.IGNORECASE)
-END_PATTERN = re.compile(r"""end;""", re.IGNORECASE)
+BEGIN_PATTERN = re.compile(r"""begin (\w+)(\s*|\[.*\]);""", 
+re.IGNORECASE)
+END_PATTERN = re.compile(r"""end\s*;""", re.IGNORECASE)
 NTAX_PATTERN = re.compile(r"""ntax=(\d+)""", re.IGNORECASE)
 NCHAR_PATTERN = re.compile(r"""nchar=(\d+)""", re.IGNORECASE)
 COMMENT_PATTERN = re.compile(r"""(\[.*?\])""")
@@ -29,11 +31,7 @@ MESQUITE_LINK_PATTERN = re.compile(
 
 class NexusFormatException(Exception):
     """Generic Exception for Nexus Format Errors"""
-    def __init__(self, arg):
-        self.value = arg
-    
-    def __str__(self):
-        return repr(self.value)
+    pass
 
 
 class GenericHandler(object):
@@ -127,23 +125,38 @@ class TaxaHandler(GenericHandler):
         """
         super(TaxaHandler, self).parse(data)
         in_taxlabel_block = False
+        found_ntaxa = None
         for line in data:
             line = self.remove_comments(line).strip()
             line = QUOTED_PATTERN.sub('\\1', line)
             if self.is_dimensions.match(line):
                 found_ntaxa = int(self.is_dimensions.findall(line)[0])
+            elif 'begin taxa' in line.lower():
+                continue
             elif line == ';':
                 continue
-            elif self.is_taxlabel_block.match(line):
+            elif in_taxlabel_block or self.is_taxlabel_block.match(line):
                 in_taxlabel_block = True
-            elif in_taxlabel_block:
-                for taxon in line.strip(';').split():
-                    self.taxa.append(taxon)
+                line = self.is_taxlabel_block.sub("", line)
+                taxa = [t.replace(";", "").strip() for t in line.split(" ")]
+                taxa = [t for t in taxa if len(t) and t not in self.taxa]
+                if len(taxa):
+                    self.taxa.extend(taxa)
             elif MESQUITE_TITLE_PATTERN.match(line):
                 self.attributes.append(line)
             elif MESQUITE_LINK_PATTERN.match(line):
                 self.attributes.append(line)
-        assert found_ntaxa == self.ntaxa
+            else:
+                taxa = [t.replace(";", "").strip() for t in line.split(" ")]
+                taxa = [t for t in taxa if len(t) and t not in self.taxa]
+                if len(taxa):
+                    self.taxa.extend(taxa)
+        
+        if found_ntaxa and found_ntaxa != self.ntaxa:
+            raise NexusFormatException(
+                "Number of taxa (%d) doesn't match dimensions declaration (%d)" % (self.ntaxa, found_ntaxa)
+            )
+
 
     def write(self):
         """
@@ -234,18 +247,20 @@ class TreeHandler(GenericHandler):
                 if translation_pattern.match(line):
                     taxon_id, taxon = translation_pattern.findall(line)[0]
                     taxon = taxon.strip("'")
-                    assert taxon_id not in self.translators, \
-                         "Duplicate Taxa ID %s in translate block" % taxon_id
+                    if taxon_id in self.translators:
+                        raise NexusFormatException(
+                            "Duplicate Taxa ID %s in translate block" % taxon_id
+                        )
+                    if taxon in self.translators.values():
+                        raise NexusFormatException(
+                            "Duplicate Taxon %s in translate block" % taxon
+                        )
+                        
                     self.translators[taxon_id] = taxon
                 if line.endswith(';'):
                     lost_in_translation = False
 
             elif self.is_tree.search(line):
-                if lost_in_translation:
-                    # can't find a tree if we're still in the translate block!
-                    raise NexusFormatException(
-                        "Tree block has incomplete translate table"
-                    )
                 self.trees.append(line)
 
         # get taxa if not translated.
@@ -331,7 +346,9 @@ class TreeHandler(GenericHandler):
             for index in sorted([int(k) for k in self.translators.keys()]):
                 out.append("\t%d %s," % (index, self.translators[str(index)]))
             # handle last taxa label in translate block
-            out[-1] = out[-1].replace(',', ';')
+            out[-1] = out[-1].replace(',', '')
+            # work around bug https://github.com/CompEvol/beast2/issues/713
+            out.append(';')
         for tree in self.trees:
             out.append("\t" + tree)
         out.append('end;\n')
@@ -457,31 +474,36 @@ class DataHandler(GenericHandler):
         :raises NexusFormatException: If data matrix contains incomplete
             multistate values
         """
-        out = []
-        multistate = False
-        sites = [site for site in sites]
-        while len(sites) > 0:
-            site = sites.pop(0)
-            if site == ' ':
-                continue
-            elif site == '(':
-                # read-ahead
-                site = ''  # discard open bracket
-                multistate = True
-                while multistate:
-                    nextchar = sites.pop(0)
-                    if nextchar == ')':
-                        multistate = False
-                    else:
-                        site += nextchar
-            out.append(site)
+        remove_values = [' ', ';']
+        sites = [s for s in sites if s not in remove_values]
+        if '(' not in sites:
+            return sites
+        else:
+            multistate = False
+            out = []
+            # Slow reader for multistate
+            while len(sites) > 0:
+                site = sites.pop(0)
+                if site == ',':
+                    continue
+                elif site == '(':
+                    # read-ahead
+                    site = ''  # discard open bracket
+                    multistate = True
+                    while multistate:
+                        nextchar = sites.pop(0)
+                        if nextchar == ')':
+                            multistate = False
+                        else:
+                            site += nextchar
+                out.append(site)
             
-        # check we're not in hanging multistate chunk
-        if multistate:
-            raise NexusFormatException(
-                "Data Matrix contains incomplete multistate values"
-            )
-        return out
+            # check we're not in hanging multistate chunk
+            if multistate:
+                raise NexusFormatException(
+                    "Data Matrix contains incomplete multistate values"
+                )
+            return out
 
     def add_taxon(self, taxon, site_values=None):
         """
@@ -619,7 +641,7 @@ class DataHandler(GenericHandler):
                     # Datatype must come first!
                     fstring.insert(1, "%s=%s" % (key, value))
                 elif key in ('interleave', ):
-                    # IGNORE the interleaving
+                    # IGNORE the interleaving -- not implemented
                     continue
                 else:
                     if key == 'symbols':
@@ -652,6 +674,10 @@ class DataHandler(GenericHandler):
             (self.nchar, self.ntaxa)
 
 
+class CharacterHandler(DataHandler):
+    pass
+
+
 class NexusReader(object):
     """A nexus reader"""
     def __init__(self, filename=None, debug=False):
@@ -660,7 +686,7 @@ class NexusReader(object):
         self.raw_blocks = {}
         self.handlers = {
             'data': DataHandler,
-            'characters': DataHandler,
+            'characters': CharacterHandler,
             'trees': TreeHandler,
             'taxa': TaxaHandler,
         }
@@ -670,12 +696,15 @@ class NexusReader(object):
     def _do_blocks(self):
         """Iterates over all nexus blocks and parses them appropriately"""
         for block, data in self.raw_blocks.items():
-            if block == 'characters':
-                block = 'data'  # override
             self.blocks[block] = self.handlers.get(block, GenericHandler)()
             self.blocks[block].parse(data)
+        
+        if self.blocks.get('characters') and not self.blocks.get('data'):
+            self.blocks['data'] = self.blocks['characters']
+        
+        for block in self.blocks:
             setattr(self, block, self.blocks[block])
-
+        
     def read_file(self, filename):
         """
         Loads and Parses a Nexus File
@@ -694,10 +723,9 @@ class NexusReader(object):
             raise IOError("Unable To Read File %s" % filename)
 
         if filename.endswith('.gz'):
-            import gzip
-            handle = gzip.open(filename, 'rb')
+            handle = gzip.open(filename, 'rb') # pragma: no cover
         else:
-            handle = open(filename, 'rU')
+            handle = open(filename, 'r')
         self._read(handle)
         handle.close()
 
@@ -731,9 +759,9 @@ class NexusReader(object):
             # check if we're in a block and initialise
             found = BEGIN_PATTERN.findall(line)
             if found:
-                block = found[0].lower()
+                block = found[0][0].lower()
                 if block in store:
-                    raise Exception("Duplicate Block %s" % block)
+                    raise NexusFormatException("Duplicate Block %s" % block)
                 store[block] = []
 
             # check if we're ending a block
@@ -768,7 +796,3 @@ class NexusReader(object):
         handle = open(filename, 'w')
         handle.writelines(self.write())
         handle.close()
-
-
-if __name__ == '__main__':
-    pass
